from __future__ import annotations

from abc import ABC, abstractmethod

from c3 import consts
from c3.utils import C3DateTime


class RTLogRecord(ABC):
    @abstractmethod
    def is_door_alarm(self) -> bool:
        ...

    @abstractmethod
    def is_event(self) -> bool:
        ...


class DoorAlarmStatusRecord(RTLogRecord):
    """Realtime Log record for a door and alarm status"""
    def __init__(self):
        self.alarm_status = bytes(4)
        self.dss_status = bytes(4)
        self.verified: consts.VerificationMode = consts.VerificationMode.NONE
        self.event_type: consts.EventType = consts.EventType.NA
        self.time_second = 0

    @classmethod
    def from_bytes(cls, data: bytes):
        """Create DoorAlarmStatusRecord from binary log
        An RTLog is a binary message of 16 bytes send by the C3 access panel.
        If the value of byte 10 (the event type) is 255, the RTLog is a Door/Alarm Realtime Status.
        If the value of byte 10 (the event type) is not 255, the RTLog is a Realtime Event.

        Door/Alarm Realtime Status record
        All multibyte values are stored as Little-endian.

        Byte:                    0  1  2  3  4  5  6  7  8  9  A  B  C  D  E  F
                                 01:4f:86:00:99:92:98:00:04:01:00:00:a5:ad:ad:21
        Alarm status (byte 4-7):             |
                                             99:92:98:00 => (big endian:) 00989299 = 9999001
        DSS status (byte 0-3):   |
                                 01:4f:86:00 => (big endian:) 00864f01 = 8802049
        Verified (byte 8):                               |
                                                         04
        Unused (byte 9):                                    |
                                                            01
        EventType (byte 10):                                   |
                                                               00
        Unused (byte 11):                                         |
                                                                  00
                                                                     |
        Time_second (byte 12-15)                                     a5:ad:ad:21 => (big endian:) 21ADADA5 =
                                                                                                  2017-7-30 16:51:49
        """
        record = DoorAlarmStatusRecord()
        record.alarm_status = bytes(data[0:4])
        record.dss_status = bytes(data[4:8])
        try:
            record.verified = consts.VerificationMode(data[9])
        except ValueError:
            record.verified = consts.VerificationMode.OTHER
        try:
            record.event_type = consts.EventType(data[10])
        except ValueError:
            record.event_type = consts.EventType.UNKNOWN_UNSUPPORTED
        record.time_second = C3DateTime.from_value(int.from_bytes(data[12:16], byteorder="little"))
        return record

    @classmethod
    def from_kv(cls, data: dict):
        """Create EventRecord from text-based key/value log
        A key/value log contains the following fields:
        {
          'time': '2023-12-09 15:09:33',
          'sensor': '24',
          'relay': '04',
          'alarm': '00000000'
        }
        """
        record = DoorAlarmStatusRecord()
        record.alarm_status = bytes.fromhex(data['alarm'])
        sensor = int(data['sensor'], base=16)
        record.dss_status = bytes([(sensor >> (i*2)) & 0x03 for i in range(0, 4)])
        # relay = 00
        # try:
        #     record.verified = consts.VerificationMode(data[9])
        # except ValueError:
        #     record.verified = consts.VerificationMode.OTHER
        # try:
        #     record.event_type = consts.EventType(data[10])
        # except ValueError:
        #     record.event_type = consts.EventType.UNKNOWN_UNSUPPORTED
        record.time_second = C3DateTime.from_str(data['time'])
        return record

    def is_door_alarm(self) -> bool:
        return True

    def is_event(self) -> bool:
        return False

    def get_alarms(self, door_nr: int) -> list[consts.AlarmStatus]:
        alarms = []

        for i in range(0, 3):
            if i+1 == door_nr or not door_nr:
                if self.alarm_status[i] & consts.AlarmStatus.ALARM:
                    if alarms.count(consts.AlarmStatus.ALARM) == 0:
                        alarms.append(consts.AlarmStatus.ALARM)
                elif self.alarm_status[i] & consts.AlarmStatus.DOOR_OPEN_TIMEOUT:
                    if alarms.count(consts.AlarmStatus.DOOR_OPEN_TIMEOUT) == 0:
                        alarms.append(consts.AlarmStatus.DOOR_OPEN_TIMEOUT)

        return alarms

    def has_alarm(self, door_nr: int, status: consts.AlarmStatus = None):
        return ((self.alarm_status[door_nr-1] & (status or 0)) == status) or \
               ((self.alarm_status[door_nr-1] > 0) and status is None)

    def door_sensor_status(self, door_nr: int) -> consts.InOutStatus:
        return consts.InOutStatus(self.dss_status[door_nr - 1] & 0x0F)

    def door_is_open(self, door_nr: int):
        is_open = None

        if self.door_sensor_status(door_nr) == consts.InOutStatus.OPEN:
            is_open = True
        elif self.door_sensor_status(door_nr) == consts.InOutStatus.CLOSED:
            is_open = False

        return is_open

    def __repr__(self):
        repr_arr = ["Door/Alarm Realtime Status:",
                    "%-12s %-10s" % ("time_second", self.time_second),
                    "%-12s %-10s %s" % ("event_type", self.event_type, repr(self.event_type)),
                    "%-12s %-10s %s" % ("verified", self.verified, repr(self.verified)),
                    "%-12s %-10s" % ("alarm_status", self.alarm_status.hex())]

        for i in range(0, 4):
            for status in consts.AlarmStatus:
                if status != consts.AlarmStatus.NONE:
                    if self.alarm_status[i] & status == status:
                        repr_arr.append("    Door %-2s %-4s %s" % (i, status, repr(status)))

        repr_arr.append("%-12s %-10s" % ("dss_status", self.dss_status.hex()))
        for i in range(0, 4):
            repr_arr.append("    Door %-2s %-4s %s" % (i+1, self.dss_status[i],
                                                       repr(consts.InOutStatus(self.dss_status[i] & 0x0F))))

        return "\r\n".join(repr_arr)


class EventRecord(RTLogRecord):
    """Realtime Event record"""
    def __init__(self):
        self.card_no = 0
        self.pin = 0
        self.verified: consts.VerificationMode = consts.VerificationMode.NONE
        self.port_nr: int = 0
        self.event_type: consts.EventType = consts.EventType.NA
        self.in_out_state: consts.InOutDirection = consts.InOutDirection.NONE
        self.time_second = 0

    @classmethod
    def from_bytes(cls, data: bytes):
        """Create EventRecord from binary log
        All multibyte values are stored as Little-endian.

        Byte:              0  1  2  3  4  5  6  7  8  9  A  B  C  D  E  F
                           01:4f:86:00:99:92:98:00:04:01:00:00:a5:ad:ad:21
        Cardno (byte 4-7):             |
                                       99:92:98:00 => (big endian:) 00989299 = 9999001
        Pin (byte 0-3):    |
                           01:4f:86:00 => (big endian:) 00864f01 = 8802049
        Verified (byte 8):                         |
                                                   04
        DoorID (byte 9):                              |
                                                      01
        EventType (byte 10):                             |
                                                         00
        InOutState (byte 11):                               |
                                                            00
                                                               |
        Time_second (byte 12-15)                               a5:ad:ad:21 => (big endian)21ADADA5 = 2017-07-30 16:51:49
        """
        record = EventRecord()
        record.card_no = int.from_bytes(data[0:4], byteorder="little")
        record.pin = int.from_bytes(data[4:8], byteorder="little")
        try:
            record.verified = consts.VerificationMode(data[8])
        except ValueError:
            record.verified = consts.VerificationMode.OTHER
        record.port_nr = data[9]
        try:
            record.event_type = consts.EventType(data[10])
        except ValueError:
            record.event_type = consts.EventType.UNKNOWN_UNSUPPORTED
        try:
            record.in_out_state = consts.InOutDirection(data[11])
        except ValueError:
            record.in_out_state = consts.InOutDirection.UNKNOWN_UNSUPPORTED
        record.time_second = C3DateTime.from_value(int.from_bytes(data[12:16], byteorder="little"))
        return record

    @classmethod
    def from_kv(cls, data: dict):
        """Create EventRecord from text-based key/value log
        A key/value log contains the following fields:
        {
          'time': '2023-12-06 22:33:15',
          'pin': '0',
          'cardno': '0',
          'eventaddr': '1',
          'event': '8',
          'inoutstatus': '2',
          'verifytype': '200',
          'index': '9'
        }
        """
        record = EventRecord()
        record.card_no = int(data['cardno'])
        record.pin = int(data['pin'])
        try:
            record.verified = consts.VerificationMode(int(data['verifytype']))
        except ValueError:
            record.verified = consts.VerificationMode.OTHER
        record.port_nr = int(data['eventaddr'])
        try:
            record.event_type = consts.EventType(int(data['event']))
        except ValueError:
            record.event_type = consts.EventType.UNKNOWN_UNSUPPORTED
        try:
            record.in_out_state = consts.InOutDirection(int(data['inoutstatus']))
        except ValueError:
            record.in_out_state = consts.InOutDirection.UNKNOWN_UNSUPPORTED
        record.time_second = C3DateTime.from_str(data['time'])
        return record

    def is_door_alarm(self) -> bool:
        return False

    def is_event(self) -> bool:
        return True

    def __repr__(self):
        repr_arr = ["Realtime Event:",
                    "%-12s %-10s" % ("time_second", self.time_second),
                    "%-12s %-10s %s" % ("event_type", self.event_type, repr(self.event_type)),
                    "%-12s %-10s %s" % ("in_out_state", self.in_out_state, repr(self.in_out_state)),
                    "%-12s %-10s %s" % ("verified", self.verified, repr(self.verified)),
                    "%-12s %-10s" % ("card_no", self.card_no),
                    # "%-12s %-10s" % ("pin", self.pin),
                    "%-12s %-10s" % ("port_no", self.port_nr)]

        return "\r\n".join(repr_arr)


def factory(log_message: [bytes, bytearray, dict]) -> DoorAlarmStatusRecord | EventRecord:
    if isinstance(log_message, (bytes, bytearray)):
        if log_message[10] == consts.EventType.DOOR_ALARM_STATUS:
            rtlog = DoorAlarmStatusRecord.from_bytes(log_message)
        else:
            rtlog = EventRecord.from_bytes(log_message)
    elif isinstance(log_message, dict):
        if "event" in log_message:
            rtlog = EventRecord.from_kv(log_message)
        else:
            rtlog = DoorAlarmStatusRecord.from_kv(log_message)
    else:
        raise NotImplementedError(f"RT log type {type(log_message)} is not supported")

    return rtlog



from unittest import mock

from c3.core import C3
from c3.consts import VerificationMode, InOutStatus, EventType
from c3.rtlog import EventRecord, DoorAlarmStatusRecord
from c3.utils import C3DateTime


def test_c3_rtlog_event_decode1():
    raw_data = bytes([0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xc8, 0x01, 0x66, 0x02, 0x32, 0x8f, 0xae, 0x21])
    # assert 1501491250 == int.from_bytes(raw_data[12:16], byteorder="little") #11439922
    # assert 1501491250 == int.from_bytes(raw_data[12:16], byteorder="big") #3313582
    # assert 1501491250 == int.from_bytes(reversed(raw_data[12:16]), byteorder="little") # 3313582
    # assert 1501491250 == int.from_bytes(reversed(raw_data[12:16]), byteorder="big") #11439922
    record = EventRecord.from_bytes(raw_data)
    # assert C3DateTime.from_value(1501491250) == record.time_second
    assert C3DateTime(year=2017, month=7, day=31, hour=8, minute=54, second=10) == record.time_second
    assert not record.is_door_alarm()
    assert record.is_event()
    print(repr(record))


def test_c3_rtlog_event_decode2():
    raw_data = bytes([0x17, 0x30, 0x64, 0x12, 0xe2, 0xb1, 0x04, 0x00, 0x04, 0x01, 0x00, 0x00, 0x74, 0x2c, 0xaf, 0x21])
    record = EventRecord.from_bytes(raw_data)
    # assert C3DateTime.from_value(1501531508) == record.time_second
    assert not record.is_door_alarm()
    assert record.is_event()
    print(repr(record))


def test_c3_rtlog_decode_status1():
    raw_data = bytes([0x03, 0x00, 0x00, 0x00, 0x11, 0x00, 0x00, 0x00, 0x00, 0x01, 0xff, 0x00, 0xf2, 0x31, 0xb3, 0x21])
    record = DoorAlarmStatusRecord.from_bytes(raw_data)
    assert 0 < len(record.get_alarms(door_nr=None))
    assert record.has_alarm(1)
    assert not record.has_alarm(2)
    assert record.has_alarm(1, 1)
    assert record.has_alarm(1, 2)
    # assert not record.has_alarm(1, 3)
    assert False is record.door_is_open(1)
    assert None is record.door_is_open(3)
    assert record.is_door_alarm()
    assert not record.is_event()
    print(repr(record))


def test_c3_rtlog_decode_status2():
    raw_data = bytes([0x03, 0x00, 0x00, 0x00, 0x02, 0x00, 0x00, 0x00, 0x01, 0x01, 0xff, 0x00, 0xfc, 0x31, 0xb3, 0x21])
    record = DoorAlarmStatusRecord.from_bytes(raw_data)
    assert record.has_alarm(1)
    assert record.door_is_open(1)
    assert None is record.door_is_open(2)
    assert record.is_door_alarm()
    assert not record.is_event()
    print(repr(record))


@mock.patch.object(C3, '_update_inout_status', new_callable=mock.MagicMock)
def test_rtlog_unknown_verification_mode(_unused_update_inout_status_mock):
    with mock.patch('socket.socket') as mock_socket:
        panel = C3('localhost')
        mock_socket.return_value.send.return_value = 8
        mock_socket.return_value.recv.side_effect = [
            bytes.fromhex("aa00c80400"), bytes.fromhex("d805fefea30955"),
            bytes.fromhex("aa00c80400"), bytes.fromhex("d805000062e955")
        ]

        assert panel.connect() is True

        mock_socket.return_value.recv.side_effect = [
            bytes.fromhex("aa01c81400"),
            bytes.fromhex("d805e8ff0200000011000000000bff00fa61fb2c38b255")]
        logs = panel.get_rt_log()
        assert len(logs) == 1
        assert isinstance(logs[0], DoorAlarmStatusRecord)
        assert logs[0].verified == VerificationMode.CARD_WITH_PASSWORD

        mock_socket.return_value.recv.side_effect = [
            bytes.fromhex("aa01c81400"),
            bytes.fromhex("d805e8ff02000000110007000007ff00fa61fb2c456855")]
        logs = panel.get_rt_log()
        assert len(logs) == 1
        assert logs[0].verified == VerificationMode.OTHER


@mock.patch.object(C3, '_update_inout_status', new_callable=mock.MagicMock)
def test_rtlog_keyvalue_response(_unused_update_inout_status_mock):
    with mock.patch('socket.socket') as mock_socket:
        panel = C3('localhost')
        mock_socket.return_value.send.return_value = 8
        mock_socket.return_value.recv.side_effect = [
            bytes.fromhex("aa02c90100"), bytes.fromhex("f357d955"),
            bytes.fromhex("aa02c80000"), bytes.fromhex("81fe55"),
            bytes.fromhex("aa02c84200"), bytes.fromhex("7e53657269616c4e756d6265723d4143"
                                                       "59543033323335333632372c4c6f636b"
                                                       "436f756e743d342c417578496e436f756"
                                                       "e743d342c4175784f7574436f756e743d"
                                                       "34599355"),
            # Not sure what the message structure of the response use protocol version (?) 02 is
            bytes.fromhex("aa01c86801"), bytes.fromhex("000000000000000000000000000000000000000000000000000000"
                                                       "0000000000000000000000000000000000000000000000000000000000000000"
                                                       "0000000000000000000000000000000000000000000000000000000000000000"
                                                       "0000000000000000000000000000000000000000000000000000000000000000"
                                                       "0000000000000000000000000000000000000000000000000000000000000000"
                                                       "0000000000000000000000000000000000000000000000000000000000000000"
                                                       "0000000000000000000000000000000000000000000000000000000000000000"
                                                       "0000000000000000000000000000000000000000000000000000000000000000"
                                                       "0000000000000000000000000000000000000000000000000000000000000000"
                                                       "0000000000000000000000000000000000000000000000000000000000000000"
                                                       "0000000000000000000000000000000000000000000000000000000000000000"
                                                       "00000000000000000000000000fabb55"),
            bytes.fromhex("aa01c8c400"), bytes.fromhex("74696d653d323032332d31322d30362032323a33333a3135097069"
                                                       "6e3d3009636172646e6f3d30096576656e74616464723d30096576656e743d32"
                                                       "303609696e6f75747374617475733d3209766572696679747970653d32303009"
                                                       "696e6465783d380d0a74696d653d323032332d31322d30362032323a35373a35"
                                                       "340970696e3d3009636172646e6f3d30096576656e74616464723d3109657665"
                                                       "6e743d3809696e6f75747374617475733d3209766572696679747970653d3230"
                                                       "3009696e6465783d39621455"),
            bytes.fromhex("aa01c83a00"), bytes.fromhex("74696d653d323032332d31322d30392031353a30393a33330973656e736f72"
                                                       "3d32340972656c61793d303409616c61726d3d3030303030303030c2b655")
        ]

        assert panel.connect() is True

        logs = panel.get_rt_log()
        assert len(logs) == 0

        logs = panel.get_rt_log()
        assert len(logs) == 1
        assert isinstance(logs[0], EventRecord)
        assert logs[0].port_nr == 1
        assert logs[0].event_type == EventType.REMOTE_OPENING
        assert logs[0].verified == VerificationMode.OTHER

        logs = panel.get_rt_log()
        assert len(logs) == 1
        assert isinstance(logs[0], DoorAlarmStatusRecord)
        assert not logs[0].door_is_open(1)
        assert not logs[0].door_is_open(2)
        assert logs[0].door_is_open(3)
        assert not logs[0].door_is_open(4)
        assert logs[0].door_sensor_status(1) == InOutStatus.UNKNOWN
        assert logs[0].door_sensor_status(2) == InOutStatus.CLOSED
        assert logs[0].door_sensor_status(3) == InOutStatus.OPEN
        assert logs[0].door_sensor_status(4) == InOutStatus.UNKNOWN
        assert not logs[0].get_alarms(1)
        assert not logs[0].get_alarms(2)
        assert not logs[0].get_alarms(3)
        assert not logs[0].get_alarms(4)



from c3.utils import *


def test_lsb():
    assert 0xaa == lsb(0x123aa)
    assert 0 == lsb(0xAB00)


def test_msb():
    assert 0 == msb(0xEF)
    assert 0xcd == msb(0x12cd34)


def test_c3_datetime_from_value():
    c3_dt = C3DateTime.from_value(347748895)
    assert 2010 == c3_dt.year
    assert 10 == c3_dt.month
    assert 54 == c3_dt.minute

    assert C3DateTime(year=2017, month=7, day=30, hour=15, minute=24, second=32) == \
           C3DateTime.from_value(int.from_bytes([0x21, 0xad, 0x99, 0x30], byteorder="big"))
    assert C3DateTime(year=2013, month=10, day=8, hour=14, minute=38, second=32) == \
           C3DateTime.from_value(int.from_bytes(reversed([0x1a, 0x61, 0x70, 0xe8]), byteorder="little"))


def test_c3_datetime_to_value():
    c3_dt = C3DateTime(2010, 10, 26, 20, 54, 55)
    assert 347748895 == c3_dt.to_value()



from c3 import controldevice
from c3 import consts


def test_c3_device_control_message_output_door():
    output_operation = controldevice.ControlDeviceOutput(1, consts.ControlOutputAddress.DOOR_OUTPUT, 200)
    assert bytes([0x01, 0x01, 0x01, 0xc8, 0x00]) == output_operation.to_bytes()
    print(repr(output_operation))


def test_c3_device_control_message_output_door_int():
    output_operation = controldevice.ControlDeviceOutput(3, 1, 200)
    assert bytes([0x01, 0x03, 0x01, 0xc8, 0x00]) == output_operation.to_bytes()
    print(repr(output_operation))


def test_c3_device_control_message_output_aux():
    output_operation = controldevice.ControlDeviceOutput(2, consts.ControlOutputAddress.AUX_OUTPUT, 100)
    assert bytes([0x01, 0x02, 0x02, 0x64, 0x00]) == output_operation.to_bytes()
    print(repr(output_operation))


def test_c3_device_control_message_cancel():
    output_operation = controldevice.ControlDeviceCancelAlarms()
    assert bytes([0x02, 0x00, 0x00, 0x00, 0x00]) == output_operation.to_bytes()
    print(repr(output_operation))


def test_c3_device_control_message_restart():
    output_operation = controldevice.ControlDeviceRestart()
    assert bytes([0x03, 0x00, 0x00, 0x00, 0x00]) == output_operation.to_bytes()
    print(repr(output_operation))


def test_c3_device_control_message_nostate():
    output_operation = controldevice.ControlDeviceNormalOpenStateEnable(2, 1)
    assert bytes([0x04, 0x02, 0x01, 0x00, 0x00]) == output_operation.to_bytes()



from enum import IntEnum, unique
from collections import namedtuple

# Defaults
C3_PORT_DEFAULT = 4370
C3_PORT_BROADCAST = 65535

# Protocol commands
C3_MESSAGE_START = 0xAA
C3_MESSAGE_END = 0x55
C3_PROTOCOL_VERSION = 0x01
C3_DISCOVERY_MESSAGE = "CallSecurityDevice"


class Command(IntEnum):
    """Enumeration of supported device_name interaction commands"""
    DISCOVER = 0x14
    CONNECT_SESSION = 0x76
    CONNECT_SESSION_LESS = 0x01
    DISCONNECT = 0x02
    GETPARAM = 0x04
    DATATABLE_CFG = 0x06
    CONTROL = 0x05
    RTLOG_BINARY = 0x0B
    RTLOG_KEYVALUE = 0x79


C3_REPLY_OK = 0xC8
C3_REPLY_ERROR = 0xC9


Errors = {
    -13: "Command error: This command is not available",
    -14: "The communication password is not correct",
}


@unique
class _IntEnumWithDescription(IntEnum):
    def __new__(cls, *args):
        obj = int.__new__(cls, args[0])
        obj._value_ = args[0]
        return obj

    # ignore the first param since it's already set by __new__
    def __init__(self, _: str, description: str = None):
        self._description_ = description

    def __str__(self):
        return str(self.value)

    def __repr__(self):
        return self._description_

    # this makes sure that the description is read-only
    @property
    def description(self):
        return self._description_


# Control operations
class ControlOperation(_IntEnumWithDescription):
    OUTPUT = 1,         "Output operation (door or auxiliary)"
    CANCEL_ALARM = 2,   "Cancel alarm"
    RESTART_DEVICE = 3, "Restart Device"
    ENDIS_NO_STATE = 4, "Enable/disable normal open state"


class ControlOutputAddress(_IntEnumWithDescription):
    DOOR_OUTPUT = 1, "Door output"
    AUX_OUTPUT = 2,  "Auxiliary output"


class DoorSensorType(_IntEnumWithDescription):
    NONE = 0,         "Not available"
    NORMAL_OPEN = 1,  "Normal open"
    NORMAL_CLOSE = 2, "Normal close"


# Event values
class VerificationMode(_IntEnumWithDescription):
    NONE = 0,                "None"
    FINGER = 1,              "Only finger"
    PASSWORD = 3,            "Only password"
    CARD = 4,                "Only card"
    CARD_OR_FINGER = 6,      "Card or finger"
    CARD_WITH_FINGER = 10,   "Card and finger"
    CARD_WITH_PASSWORD = 11, "Card and password"
    OTHER = 200,             "Others"


class EventType(_IntEnumWithDescription):
    NA = -1, "N/A"
    NORMAL_PUNCH_OPEN = 0, "Normal Punch Open"
    PUNCH_NORMAL_OPEN_TZ = 1, "Punch during Normal Open Time Zone"
    FIRST_CARD_NORMAL_OPEN = 2, "First Card Normal Open (Punch Card)"
    MULTI_CARD_OPEN = 3, "Multi-Card Open (Punching Card)"
    EMERGENCY_PASS_OPEN = 4, "Emergency Password Open"
    OPEN_NORMAL_OPEN_TZ = 5, "Open during Normal Open Time Zone"
    LINKAGE_EVENT_TRIGGER = 6, "Linkage Event Triggered"
    CANCEL_ALARM = 7, "Cancel Alarm"
    REMOTE_OPENING = 8, "Remote Opening"
    REMOTE_CLOSING = 9, "Remote Closing"
    DISABLE_INTRADAY_NORMAL_OPEN_TZ = 10, "Disable Intraday Normal Open Time Zone"
    ENABLE_INTRADAY_NORMAL_OPEN_TZ = 11, "Enable Intraday Normal Open Time Zone"
    OPEN_AUX_OUTPUT = 12, "Open Auxiliary Output"
    CLOSE_AUX_OUTPUT = 13, "Close Auxiliary Output"
    PRESS_FINGER_OPEN = 14, "Press Fingerprint Open"
    MULTI_CARD_OPEN_FP = 15, "Multi-Card Open (Press Fingerprint)"
    FP_NORMAL_OPEN_TZ = 16, "Press Fingerprint during Normal Open Time Zone"
    CARD_FP_OPEN = 17, "Card plus Fingerprint Open"
    FIRST_CARD_NORMAL_OPEN_FP = 18, "First Card Normal Open (Press Fingerprint)"
    FIRST_CARD_NORMAL_OPEN_CARD_FP = 19, "First Card Normal Open (Card plus Fingerprint)"
    TOO_SHORT_PUNCH_INTERVAL = 20, "Too Short Punch Interval"
    DOOR_INACTIVE_TZ = 21, "Door Inactive Time Zone (Punch Card)"
    ILLEGAL_TZ = 22, "Illegal Time Zone"
    ACCESS_DENIED = 23, "Access Denied"
    ANTI_PASSBACK = 24, "Anti-Passback"
    INTERLOCK = 25, "Interlock"
    MULTI_CARD_AUTH = 26, "Multi-Card Authentication (Punching Card)"
    UNREGISTERED_CARD = 27, "Unregistered Card"
    OPENING_TIMEOUT = 28, "Opening Timeout:"
    CARD_EXPIRED = 29, "Card Expired"
    PASSWORD_ERROR = 30, "Password Error"
    TOO_SHORT_FP_INTERVAL = 31, "Too Short Fingerprint Pressing Interval"
    MULTI_CARD_AUTH_FP = 32, "Multi-Card Authentication (Press Fingerprint)"
    FP_EXPIRED = 33, "Fingerprint Expired"
    UNREGISTERED_FP = 34, "Unregistered Fingerprint"
    DOOR_INACTIVE_TZ_FP = 35, "Door Inactive Time Zone (Press Fingerprint)"
    DOOR_INACTIVE_TZ_EXIT = 36, "Door Inactive Time Zone (Exit Button)"
    FAILED_CLOSE_NORMAL_OPEN_TZ = 37, "Failed to Close during Normal Open Time Zone"
    VERIFY_TYPE_INVALID = 41, "Verify Type Invalid"
    WG_FORMAT_ERROR = 42, "WG Format Error"
    DURESS_PASSWORD_OPEN = 101, "Duress Password Open"
    OPENED_ACCIDENT = 102, "Opened Accidentally"
    DURESS_FP_OPEN = 103, "Duress Fingerprint Open"
    DOOR_OPENED_CORRECT = 200, "Door Opened Correctly"
    DOOR_CLOSED_CORRECT = 201, "Door Closed Correctly"
    EXIT_BUTTON_OPEN = 202, "Exit button Open"
    MULTI_CARD_OPEN_CARD_FP = 203, "Multi-Card Open (Card plus Fingerprint)"
    NORMAL_OPEN_TZ_OVER = 204, "Normal Open Time Zone Over"
    REMOTE_NORMAL_OPEN = 205, "Remote Normal Opening"
    DEVICE_START = 206, "Device Start"
    DOOR_OPEN_BY_SUPERUSER = 208, "Door Opened by Superuser"
    AUX_INPUT_DISCONNECT = 220, "Auxiliary Input Disconnected"
    AUX_INPUT_SHORT = 221, "Auxiliary Input Shorted"
    DOOR_ALARM_STATUS = 255, "Current door and alarm status"
    UNKNOWN_UNSUPPORTED = 999, "Unknown"


class InOutDirection(_IntEnumWithDescription):
    ENTRY = 0, "Entry"
    EXIT = 3,  "Exit"
    NONE = 2,  "None"
    UNKNOWN_UNSUPPORTED = 15, "Unknown"


class AlarmStatus(_IntEnumWithDescription):
    NONE = 0,              "None"
    ALARM = 1,             "Alarm"
    DOOR_OPEN_TIMEOUT = 2, "Door opening timeout"


class InOutStatus(_IntEnumWithDescription):
    UNKNOWN = 0, "Unknown"
    CLOSED = 1,  "Closed"
    OPEN = 2,    "Open"






from unittest import mock
import time
import pytest

from c3 import rtlog
from c3.core import C3
from c3 import consts
from c3 import controldevice


def test_core_init():
    panel = C3('localhost')
    assert panel.nr_of_locks == 0
    assert panel.nr_aux_in == 0
    assert panel.nr_aux_out == 0
    assert panel.lock_status(1) == consts.InOutStatus.UNKNOWN
    assert panel.aux_in_status(1) == consts.InOutStatus.UNKNOWN
    assert panel.aux_out_status(2) == consts.InOutStatus.UNKNOWN


def test_core_get_device_param():
    with mock.patch('socket.socket') as mock_socket:
        panel = C3('localhost')
        mock_socket.return_value.send.return_value = 8
        mock_socket.return_value.recv.side_effect = [
            bytes.fromhex("aa01c80400"), bytes.fromhex("d18a0000915255"),
            bytes.fromhex("aa01c80400"), bytes.fromhex("d18a000055")
        ]

        assert panel.connect() is True

        mock_socket.return_value.recv.side_effect = [bytes.fromhex("aa01c84800"),
                                                     bytes.fromhex(
            "d18a02007e5a4b465056657273696f6e3d31302c4c6f636b436f756e743d322c5265616465"
            "72436f756e743d342c417578496e436f756e743d322c4175784f7574436f756e743d32783f55")]
        params = panel.get_device_param(["~ZKFPVersion:", "LockCount", "ReaderCount", "AuxInCount", "AuxOutCount"])
        assert params["~ZKFPVersion"] == "10"
        assert params["LockCount"] == "2"
        assert params["ReaderCount"] == "4"
        assert params["AuxInCount"] == "2"
        assert params["AuxOutCount"] == "2"


def test_core_connect_response_incomplete():
    with mock.patch('socket.socket') as mock_socket:
        panel = C3('localhost')
        mock_socket.return_value.send.return_value = 8
        mock_socket.return_value.recv.side_effect = [
            bytes.fromhex("aa01c80400"), bytes.fromhex("d18a0000915255"),
            bytes.fromhex("aa01c80400"), bytes.fromhex("d18a000055")
        ]

        assert panel.connect() is True

        mock_socket.return_value.recv.side_effect = [bytes.fromhex("aa01c80800"), bytes.fromhex("d18a020012345678")]
        with pytest.raises(ValueError):
            panel.get_device_param([])


def test_core_connect_response_no_data():
    with mock.patch('socket.socket') as mock_socket:
        panel = C3('localhost')
        panel.receive_retries = 1
        mock_socket.return_value.send.return_value = 8
        mock_socket.return_value.recv.side_effect = [
            bytes.fromhex("aa01c80400"), bytes.fromhex("d18a0000915255"),
            bytes(), bytes()
        ]

        assert panel.connect() is True

        mock_socket.return_value.recv.side_effect = [bytes.fromhex("aa01c80800"), bytes()]
        with pytest.raises(ValueError):
            panel.get_device_param([])


def test_core_connect_session_less():
    with mock.patch('socket.socket') as mock_socket:
        panel = C3('localhost')
        mock_socket.return_value.send.return_value = 8
        mock_socket.return_value.recv.side_effect = [
            # Reject response to session connect attempt
            bytes.fromhex("aa01c90100"), bytes.fromhex("f313d955"),
            # Confirm session-less connection attempt
            bytes.fromhex("aa00c80000"), bytes.fromhex("81fe55"),
            # Session-less response to init-getparams
            bytes.fromhex("aa00c84800"), bytes.fromhex("7e53657269616c4e756d6265723d4444473831333030313630393232303034"
                                                       "30312c4c6f636b436f756e743d342c417578496e436f756e743d342c417578"
                                                       "4f7574436f756e743d343b3f55"),
        ]

        assert panel.connect() is True
        assert panel.serial_number == "DDG8130016092200401"
        assert panel.nr_of_locks == 4
        assert panel.nr_aux_out == 4
        assert panel.nr_aux_in == 4


def test_core_connect_password():
    with mock.patch('socket.socket') as mock_socket:
        panel = C3('localhost')
        mock_socket.return_value.send.return_value = 12
        mock_socket.return_value.recv.side_effect = [
            bytes.fromhex("aa01c80400"), bytes.fromhex("1420fefe4c5e55"),
            bytes.fromhex("aa01c80000"), bytes.fromhex("1420efe955")
        ]

        assert panel.connect("banana123") is True
        assert mock_socket.return_value.send.call_count == 2
        mock_socket.return_value.send.assert_any_call(bytes.fromhex("aa01760d00fefefefe62616E616E61313233961955"))


def test_core_lock_status():
    with mock.patch('socket.socket') as mock_socket:
        panel = C3('localhost')
        mock_socket.return_value.send.return_value = 8
        mock_socket.return_value.recv.side_effect = [bytes.fromhex("aa01c80400"),
                                                     bytes.fromhex("eb6600005c7f55"),
                                                     bytes.fromhex("aa01c84600"),
                                                     bytes.fromhex(
             "eb6601007e53657269616c4e756d6265723d363430343136323130313638392c4c6f636b"
             "436f756e743d322c417578496e436f756e743d322c4175784f7574436f756e743d326a2255")]

        assert panel.connect() is True
        assert panel.nr_of_locks == 2
        assert panel.lock_status(1) == consts.InOutStatus.UNKNOWN
        assert panel.lock_status(2) == consts.InOutStatus.UNKNOWN
        assert panel.lock_status(3) == consts.InOutStatus.UNKNOWN
        assert panel.lock_status(4) == consts.InOutStatus.UNKNOWN

        mock_socket.return_value.recv.side_effect = [bytes.fromhex("aa01c81400"),
                                                     bytes.fromhex("eb66030003000000110000000001ff00f5c1ca2caa1f55")]
        logs = panel.get_rt_log()
        assert len(logs) == 1
        assert isinstance(logs[0], rtlog.DoorAlarmStatusRecord)
        assert logs[0].door_sensor_status(1) == consts.InOutStatus.CLOSED
        assert logs[0].door_sensor_status(2) == consts.InOutStatus.UNKNOWN
        assert panel.lock_status(1) == consts.InOutStatus.CLOSED
        assert panel.lock_status(2) == consts.InOutStatus.UNKNOWN
        assert panel.lock_status(3) == consts.InOutStatus.UNKNOWN
        assert panel.lock_status(4) == consts.InOutStatus.UNKNOWN


def test_core_aux_in_status():
    with mock.patch('socket.socket') as mock_socket:
        panel = C3('localhost')
        mock_socket.return_value.send.return_value = 8
        mock_socket.return_value.recv.side_effect = [bytes.fromhex("aa01c80400"),
                                                     bytes.fromhex("eb6600005c7f55"),
                                                     bytes.fromhex("aa01c84600"),
                                                     bytes.fromhex(
             "eb6601007e53657269616c4e756d6265723d363430343136323130313638392c4c6f636b"
             "436f756e743d322c417578496e436f756e743d322c4175784f7574436f756e743d326a2255")]

        assert panel.connect() is True
        assert panel.nr_aux_in == 2
        assert panel.aux_in_status(1) == consts.InOutStatus.UNKNOWN
        assert panel.aux_in_status(2) == consts.InOutStatus.UNKNOWN

        mock_socket.return_value.recv.side_effect = [bytes.fromhex("aa01c81400"),
                                                     bytes.fromhex("eb663c000000000000000000c802dd02f5c3ca2c0abe55")]
        logs = panel.get_rt_log()
        assert len(logs) == 1
        assert isinstance(logs[0], rtlog.EventRecord)
        assert logs[0].port_nr == 2
        assert logs[0].event_type == consts.EventType.AUX_INPUT_SHORT
        assert panel.aux_in_status(1) == consts.InOutStatus.UNKNOWN
        assert panel.aux_in_status(2) == consts.InOutStatus.CLOSED

        mock_socket.return_value.recv.side_effect = [bytes.fromhex("aa01c81400"),
                                                     bytes.fromhex("eb663e000000000000000000c802dc02f9c3ca2ca98755")]
        logs = panel.get_rt_log()
        assert len(logs) == 1
        assert isinstance(logs[0], rtlog.EventRecord)
        assert logs[0].port_nr == 2
        assert logs[0].event_type == consts.EventType.AUX_INPUT_DISCONNECT
        assert panel.aux_in_status(1) == consts.InOutStatus.UNKNOWN
        assert panel.aux_in_status(2) == consts.InOutStatus.OPEN


def test_core_aux_out_status():
    with mock.patch('socket.socket') as mock_socket:
        panel = C3('localhost')
        mock_socket.return_value.send.return_value = 8
        mock_socket.return_value.recv.side_effect = [bytes.fromhex("aa01c80400"),
                                                     bytes.fromhex("eb6600005c7f55"),
                                                     bytes.fromhex("aa01c84600"),
                                                     bytes.fromhex(
             "eb6601007e53657269616c4e756d6265723d363430343136323130313638392c4c6f636b"
             "436f756e743d322c417578496e436f756e743d322c4175784f7574436f756e743d326a2255")]

        assert panel.connect() is True
        assert panel.nr_aux_out == 2
        assert panel.aux_out_status(1) == consts.InOutStatus.UNKNOWN
        assert panel.aux_out_status(2) == consts.InOutStatus.UNKNOWN

        mock_socket.return_value.recv.side_effect = [bytes.fromhex("aa01c81400"),
                                                     bytes.fromhex("eb6643000000000000000000c8020c0229c4ca2cbb6255")]
        logs = panel.get_rt_log()
        assert len(logs) == 1
        assert isinstance(logs[0], rtlog.EventRecord)
        assert logs[0].port_nr == 2
        assert logs[0].event_type == consts.EventType.OPEN_AUX_OUTPUT
        assert panel.aux_out_status(1) == consts.InOutStatus.UNKNOWN
        assert panel.aux_out_status(2) == consts.InOutStatus.OPEN

        mock_socket.return_value.recv.side_effect = [bytes.fromhex("aa01c81400"),
                                                     bytes.fromhex("eb664c000000000000000000c8020d0271c4ca2c9ac455")]
        logs = panel.get_rt_log()
        assert len(logs) == 1
        assert isinstance(logs[0], rtlog.EventRecord)
        assert logs[0].port_nr == 2
        assert logs[0].event_type == consts.EventType.CLOSE_AUX_OUTPUT
        assert panel.aux_out_status(1) == consts.InOutStatus.UNKNOWN
        assert panel.aux_out_status(2) == consts.InOutStatus.CLOSED


def test_core_aux_out_open_close():
    with mock.patch('socket.socket') as mock_socket:
        panel = C3('localhost')
        mock_socket.return_value.send.return_value = 8
        mock_socket.return_value.recv.side_effect = [bytes.fromhex("aa01c80400"),
                                                     bytes.fromhex("eb6600005c7f55"),
                                                     bytes.fromhex("aa01c84600"),
                                                     bytes.fromhex(
             "eb6601007e53657269616c4e756d6265723d363430343136323130313638392c4c6f636b"
             "436f756e743d322c417578496e436f756e743d322c4175784f7574436f756e743d326a2255")]

        assert panel.connect() is True
        assert panel.nr_aux_out == 2
        assert panel.aux_out_status(1) == consts.InOutStatus.UNKNOWN
        assert panel.aux_out_status(2) == consts.InOutStatus.UNKNOWN

        mock_socket.return_value.recv.side_effect = [bytes.fromhex("aa01c80400"),
                                                     bytes.fromhex("eb6602005d1f55")]

        command = controldevice.ControlDeviceOutput(2, consts.ControlOutputAddress.AUX_OUTPUT, 2)
        panel.control_device(command)

        mock_socket.return_value.recv.side_effect = [bytes.fromhex("aa01c81400"),
                                                     bytes.fromhex("eb6643000000000000000000c8020c0229c4ca2cbb6255")]
        panel.get_rt_log()
        assert panel.aux_out_status(1) == consts.InOutStatus.UNKNOWN
        assert panel.aux_out_status(2) == consts.InOutStatus.OPEN

        time.sleep(2.1)

        assert panel.aux_out_status(1) == consts.InOutStatus.UNKNOWN
        assert panel.aux_out_status(2) == consts.InOutStatus.CLOSED


def test_core_door_settings():
    with mock.patch('socket.socket') as mock_socket:
        panel = C3('localhost')
        mock_socket.return_value.send.return_value = 8
        mock_socket.return_value.recv.side_effect = [bytes.fromhex("aa01c80400"),
                                                     bytes.fromhex("4d9cfefe9f2655"),
                                                     bytes.fromhex("aa01c82a00"),
                                                     bytes.fromhex("4d9cfffe4c6f636b436f756e743d322c417578496e436f75"
                                                                   "6e743d322c4175784f7574436f756e743d32bb5755")]

        assert panel.connect() is True
        assert panel.nr_of_locks == 2

        mock_socket.return_value.recv.side_effect = [bytes.fromhex("aa00c83d00"),
                                                     bytes.fromhex("4d9c08ff446f6f723153656e736f72547970653d322c446f"
                                                                   "6f723144726976657274696d653d312c446f6f7231446574"
                                                                   "6563746f7274696d653d323530b18d55"),
                                                     bytes.fromhex("aa00c83c00"),
                                                     bytes.fromhex("4d9c09ff446f6f723253656e736f72547970653d302c446f"
                                                                   "6f723244726976657274696d653d352c446f6f7232446574"
                                                                   "6563746f7274696d653d3135158355")]

        assert panel.door_settings(1).sensor_type == consts.DoorSensorType.NORMAL_CLOSE
        assert panel.door_settings(1).lock_drive_time == 1
        assert panel.door_settings(1).door_alarm_timeout == 250
        assert panel.door_settings(2).sensor_type == consts.DoorSensorType.NONE
        assert panel.door_settings(2).lock_drive_time == 5
        assert panel.door_settings(2).door_alarm_timeout == 15


def test_core_update_inout_status_exit_button():
    with mock.patch('socket.socket') as mock_socket:
        panel = C3('localhost')
        mock_socket.return_value.send.return_value = 8
        mock_socket.return_value.recv.side_effect = [bytes.fromhex("aa01c80400"),
                                                     bytes.fromhex("4d9cfefe9f2655"),
                                                     bytes.fromhex("aa01c82a00"),
                                                     bytes.fromhex("4d9cfffe4c6f636b436f756e743d322c417578496e436f75"
                                                                   "6e743d322c4175784f7574436f756e743d32bb5755")]

        assert panel.connect() is True
        assert panel.nr_of_locks == 2
        assert panel.lock_status(1) == consts.InOutStatus.UNKNOWN
        assert panel.lock_status(2) == consts.InOutStatus.UNKNOWN

        mock_socket.return_value.recv.side_effect = [bytes.fromhex("aa00c81400"),
                                                     bytes.fromhex("4d9c06ff03000000111000000001ff0013ecfd2d4ae555")]

        panel.get_rt_log()
        assert panel.lock_status(1) == consts.InOutStatus.CLOSED
        assert panel.lock_status(2) == consts.InOutStatus.UNKNOWN

        mock_socket.return_value.recv.side_effect = [bytes.fromhex("aa00c81400"),
                                                     bytes.fromhex("4d9c07ff0000000000000000c802ca0215ecfd2d710f55"),
                                                     bytes.fromhex("aa00c83d00"),
                                                     bytes.fromhex("4d9c08ff446f6f723153656e736f72547970653d322c446f"
                                                                   "6f723144726976657274696d653d312c446f6f7231446574"
                                                                   "6563746f7274696d653d323530b18d55"),
                                                     bytes.fromhex("aa00c83c00"),
                                                     bytes.fromhex("4d9c09ff446f6f723253656e736f72547970653d302c446f"
                                                                   "6f723244726976657274696d653d312c446f6f7232446574"
                                                                   "6563746f7274696d653d3135507055")]

        panel.get_rt_log()
        assert panel.lock_status(1) == consts.InOutStatus.CLOSED
        assert panel.lock_status(2) == consts.InOutStatus.OPEN
        assert panel.door_settings(2).sensor_type == consts.DoorSensorType.NONE
        assert panel.door_settings(2).lock_drive_time == 1

        time.sleep(panel.door_settings(2).lock_drive_time + 0.1)

        assert panel.lock_status(2) == consts.InOutStatus.CLOSED

        mock_socket.return_value.recv.side_effect = [bytes.fromhex("aa00c81400"),
                                                     bytes.fromhex("4d9c06ff03000000111000000001ff0013ecfd2d4ae555")]

        # Subsequent DoorAlarmStatus logs with door 2 status 'unknown' are ignored,
        # last-known status better than unknown is used
        panel.get_rt_log()
        assert panel.lock_status(1) == consts.InOutStatus.CLOSED
        assert panel.lock_status(2) == consts.InOutStatus.CLOSED



from abc import ABC

from c3 import consts


class ControlDeviceBase(ABC):
    """A ControlDevice is a binary message of 5 bytes send to the C3 access panel.
    It changes the states of the doors, auxiliary relays and alarms.
    All multibyte values are stored as Little-endian.

    Byte       0  1  2  3  4
               01:01:01:c8:00
    Operation: |
               01 => 1 (1: output, 2: cancel alarm, 3: restart device_name, 4: enable/disable normal open state)
    Param 1:      |
    Param 2:         |
    Param 3:            |
    Param 4:               |

    The meaning of the parameters is depending on the Operation code.
    Param 4 is reserved for future use (defaults to 0)
    Operation 1: Output operation
      Param 1: Door number or auxiliary output number
      Param 2: The address type of output operation (1: Door output, 2: Auxiliary output)
      Param 3: Duration of the open operation, only for address type = 1 (door output).
               0: close, 255: normal open state, 1~254: normal open duration
    Operation 2: Cancel alarm
      Param 1: 0 (null)
      Param 2: 0 (null)
      Param 3: 0 (null)
    Operation 3: Restart device_name
      Param 1: 0 (null)
      Param 2: 0 (null)
      Param 3: 0 (null)
    Operation 3: Enable/disable normal open state
      Param 1: Door number
      Param 2: Enable / disable (0: disable, 1: enable)
      Param 3: 0 (null)
    """

    def __init__(self, operation: consts.ControlOperation, *args: int):
        """Constructor to initialize base class.
        The param1, param2, param3 and param4 values are provided as variable argument."""
        self.operation: consts.ControlOperation = operation
        self.param1: int = args[0] if len(args) > 0 else None
        self.param2: int = args[1] if len(args) > 1 else None
        self.param3: int = args[2] if len(args) > 2 else None
        self.param4: int = args[3] if len(args) > 3 else None

    @classmethod
    def from_bytes(cls, data: bytes):
        return ControlDeviceBase(*data)

    def to_bytes(self) -> bytes:
        return bytes([self.operation, self.param1 or 0, self.param2 or 0, self.param3 or 0, self.param4 or 0])

    def __repr__(self):
        return "\r\n".join([
            "%-12s %-10s (%s)" % ("operation", self.operation, repr(self.operation)),
            "%-12s %-10s" % ("param1", self.param1),
            "%-12s %-10s" % ("param2", self.param2),
            "%-12s %-10s" % ("param3", self.param3),
            "%-12s %-10s" % ("param4", self.param4),
        ])


class ControlDeviceOutput(ControlDeviceBase):
    def __init__(self, output_number: int, address: consts.ControlOutputAddress, duration: int):
        ControlDeviceBase.__init__(self, consts.ControlOperation.OUTPUT, output_number, address, duration)

    @property
    def output_number(self) -> int:
        return self.param1

    @property
    def address(self) -> int:
        return self.param2

    @property
    def duration(self) -> int:
        return self.param3

    def __repr__(self):
        return "\r\n".join([
            "ControlDevice Output Operation:"
            "%-12s %-10s (%s)" % ("operation", self.operation, repr(self.operation)),
            "%-12s %-10s (Door/Aux Number)" % ("param1", self.output_number),
            "%-12s %-10s %s" % ("param2", self.param2, repr(consts.ControlOutputAddress(self.address))),
            "%-12s %-10s (Duration)" % ("param3", self.duration),
        ])


class ControlDeviceCancelAlarms(ControlDeviceBase):
    def __init__(self):
        ControlDeviceBase.__init__(self, consts.ControlOperation.CANCEL_ALARM)

    def __repr__(self):
        return "\r\n".join([
            "ControlDevice Cancel Alarm Operation:",
            ControlDeviceBase.__repr__(self)
        ])


class ControlDeviceNormalOpenStateEnable(ControlDeviceBase):
    def __init__(self, door_number: int, enable: bool):
        ControlDeviceBase.__init__(self, consts.ControlOperation.ENDIS_NO_STATE, door_number, enable)

    @property
    def door_number(self) -> int:
        return self.param1

    @property
    def enabled(self) -> bool:
        return bool(self.param2)

    def __repr__(self):
        return "\r\n".join([
            "ControlDevice Normal Open State Operation:"
            "%-12s %-10s (%s)" % ("operation", self.operation, repr(self.operation)),
            "%-12s %-10s (Door Number)" % ("param1", self.door_number),
            "%-12s %-10s %s" % ("param2", self.param2, "Enable" if self.enabled else "Disable"),
        ])


class ControlDeviceRestart(ControlDeviceBase):
    def __init__(self):
        ControlDeviceBase.__init__(self, consts.ControlOperation.RESTART_DEVICE)

    def __repr__(self):
        return "\r\n".join([
            "ControlDevice Restart Operation:",
            ControlDeviceBase.__repr__(self)
        ])



from __future__ import annotations

import logging
import re
import socket
import threading
from dataclasses import dataclass, field
from typing import Dict, Optional

from c3 import consts, controldevice, crc, rtlog, utils


@dataclass
class C3DeviceInfo:
    """Basic C3 panel (connection) information, obtained from discovery"""
    host: str
    port: int = consts.C3_PORT_DEFAULT
    serial_number: str = None
    mac: str = None
    device_name: str = None
    firmware_version: str = None


@dataclass
class C3DoorSettings:
    """C3 panel door configuration and settings"""
    sensor_type: consts.DoorSensorType = consts.DoorSensorType.NONE
    """Door sensor type
    
    parameter: DoorNSensorType,
    0: Not available
    1: Normal open
    2: Normal closed 
    """
    lock_drive_time: int = None
    """Lock driver time length
    
    parameter: DoorNDrivertime,
    The value range is 0 to 255.
    0: Normal closed
    255: Normal open
    1 to 254: Door-opening duration
    """
    door_alarm_timeout: int = None
    """Timeout alarm duration of door magnet
    
    parameter: DoorNDetectortime
    The value range is 0 to 255,
    Unit: second
    """


@dataclass
class C3PanelStatus:
    """C3 panel peripheral status"""
    nr_of_locks: int = 0
    nr_aux_in: int = 0
    nr_aux_out: int = 0
    door_settings: Dict[int, C3DoorSettings] = field(default_factory=dict)
    lock_status: Dict[int, consts.InOutStatus] = field(default_factory=dict)
    aux_in_status: Dict[int, consts.InOutStatus] = field(default_factory=dict)
    aux_out_status: Dict[int, consts.InOutStatus] = field(default_factory=dict)


class C3:
    log = logging.getLogger("C3")
    log.setLevel(logging.ERROR)
    receive_timeout = 1
    receive_retries = 3

    def __init__(self, host: [str | C3DeviceInfo], port: int = consts.C3_PORT_DEFAULT) -> None:
        self._sock: socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self._sock.settimeout(self.receive_timeout)
        self._connected: bool = False
        self._session_less = False
        self._initialized = False
        self._protocol_version = None
        self._rtlog_command = consts.Command.RTLOG_BINARY
        self._session_id: int = 0xFEFE
        self._request_nr: int = -258
        self._status: C3PanelStatus = C3PanelStatus()
        if isinstance(host, C3DeviceInfo):
            self._device_info: C3DeviceInfo = host
        elif isinstance(host, str):
            self._device_info: C3DeviceInfo = C3DeviceInfo(host=host, port=port or consts.C3_PORT_DEFAULT)

    @classmethod
    def _get_message_header(cls, data: [bytes or bytearray]) -> tuple[[int or None], int, int]:
        if len(data) >= 5:
            version = data[1]
            if data[0] == consts.C3_MESSAGE_START:  # and version == consts.C3_PROTOCOL_VERSION:
                command = data[2]
                data_size = data[3] + (data[4] * 256)
            else:
                raise ValueError("Received reply does not start with start token")
        else:
            raise ValueError(f"Received reply of insufficient length {len(data)}")

        return command, data_size, version

    @classmethod
    def _get_message(cls, data: [bytes or bytearray]) -> bytearray:
        if data[-1] == consts.C3_MESSAGE_END:
            # Get the message payload, without start, crc and end bytes
            checksum = crc.crc16(data[1:-3])

            if utils.lsb(checksum) == data[-3] or utils.msb(checksum) == data[-2]:
                # Return all data without header (leading) and crc (trailing)
                message = bytearray(data[5:-3])
            else:
                raise ValueError("Payload checksum is invalid: %02x%02x expected %02x%02x" %
                                 (data[-3], data[-2], utils.lsb(checksum), utils.msb(checksum)))
        else:
            raise ValueError("Payload does not include message end marker (%02x)" % data[-1])

        return message

    @classmethod
    def _construct_message(cls, session_id: Optional[int], request_nr: Optional[int], command: consts.Command,
                           data=None) -> bytes:
        message_length = len(data or []) + (4 if (session_id and request_nr) else 0)
        message = bytearray([consts.C3_PROTOCOL_VERSION,
                             command or 0x00,
                             utils.lsb(message_length),
                             utils.msb(message_length)])
        if session_id:
            message.append(utils.lsb(session_id))
            message.append(utils.msb(session_id))
            message.append(utils.lsb(request_nr))
            message.append(utils.msb(request_nr))

        if data:
            for byte in data:
                if isinstance(byte, int):
                    message.append(byte)
                elif isinstance(byte, str):
                    message.append(ord(byte))
                else:
                    raise TypeError("Data does not contain int or str: %s is %s" % (str(byte), type(byte)))

        checksum = crc.crc16(message)
        message.append(utils.lsb(checksum))
        message.append(utils.msb(checksum))

        message.insert(0, consts.C3_MESSAGE_START)
        message.append(consts.C3_MESSAGE_END)
        return message

    def _send(self, command: consts.Command, data=None) -> int:
        message = self._construct_message(self._session_id, self._request_nr, command, data)

        self.log.debug("Sending: %s", message.hex())

        bytes_written = self._sock.send(message)
        self._request_nr = self._request_nr + 1
        return bytes_written

    def _receive(self) -> tuple[bytearray, int, int]:
        # Get the first 5 bytes
        self._sock.settimeout(self.receive_timeout)

        header = bytes()
        for _ in range(self.receive_retries):
            try:
                header = self._sock.recv(5)
                if len(header) == 5:
                    break
            except socket.timeout:
                pass

        if len(header) == 5:
            self.log.debug("Received header: %s", header.hex())

            message = bytearray()
            received_command, data_size, protocol_version = self._get_message_header(header)
            # Get the optional message data, checksum (2 bytes) and end marker (1 byte)
            payload = self._sock.recv(data_size + 3)
            if data_size > 0:
                # Process message in case data available
                self.log.debug("Receiving payload (data size %d): %s", data_size, payload.hex())
                message = self._get_message(header + payload)

            if len(message) != data_size:
                raise ValueError(f"Length of received message ({len(message)}) doesn't match specified ({data_size})")

            if received_command == consts.C3_REPLY_OK:
                pass
            elif received_command == consts.C3_REPLY_ERROR:
                error = utils.byte_to_signed_int(message[-1])
                raise ConnectionError(
                    f"Error {error} received in reply: {consts.Errors[error] if error in consts.Errors else 'Unknown'}")
        else:
            raise ConnectionError(f"Invalid response header received; expected 5 bytes, received {header}")

        return message, data_size, protocol_version

    def _send_receive(self, command: consts.Command, data=None) -> tuple[bytearray, int]:
        bytes_received = 0
        receive_data = bytearray()
        session_offset = 0

        try:
            bytes_written = self._send(command, data)
            if bytes_written > 0:
                receive_data, bytes_received, _ = self._receive()
                if not self._session_less and bytes_received > 2:
                    session_offset = 4
                    session_id = (receive_data[1] << 8) + receive_data[0]
                    # msg_seq = (receive_data[3] << 8) + receive_data[2]
                    if self._session_id != session_id:
                        raise ValueError("Data received with invalid session ID")
        except BrokenPipeError as ex:
            self._connected = False
            raise ConnectionError(f"Unexpected connection end: {ex}") from ex

        return receive_data[session_offset:], bytes_received-session_offset

    def _initialize(self):
        if not self._initialized:
            try:
                params = self.get_device_param(["~SerialNumber", "FirmVer", "DeviceName", "LockCount", "AuxInCount",
                                                "AuxOutCount"])
                self._device_info.serial_number = params.get("~SerialNumber", self._device_info.serial_number)
                self._device_info.firmware_version = params.get("FirmVer", self._device_info.firmware_version)
                self._device_info.device_name = params.get("DeviceName", self._device_info.device_name)
                self._status.nr_of_locks = int(params.get("LockCount", self._status.nr_of_locks))
                self._status.nr_aux_in = int(params.get("AuxInCount", self._status.nr_aux_in))
                self._status.nr_aux_out = int(params.get("AuxOutCount", self._status.nr_aux_out))
                self._initialized = True
            except ConnectionError as ex:
                self.log.error("Connection to %s failed: %s", self._device_info.host, ex)
            except ValueError as ex:
                self.log.error("Retrieving configuration parameters from %s failed: %s",
                               self._device_info.host, ex)

    def is_connected(self) -> bool:
        # try:
        #    # this will try to read bytes without blocking and also without removing them from buffer (peek only)
        #    data = self._sock.recv(1, socket.MSG_DONTWAIT | socket.MSG_PEEK)
        #    if len(data) == 0:
        #        return True
        # except BlockingIOError:
        #    return True  # socket is open and reading from it would block
        # except ConnectionResetError:
        #    return False  # socket was closed for some other reason
        # except Exception as e:
        #    return False
        return self._connected

    @classmethod
    def _parse_kv_from_message(cls, message: bytes) -> dict:
        kv_pairs = {}

        message_str = message.decode(encoding='ascii', errors='ignore')
        pattern = re.compile(r"([\w~]+)=([^,\t]+)")
        for (param_name, param_value) in re.findall(pattern, message_str):
            kv_pairs[param_name] = param_value

        return kv_pairs

    def __repr__(self):
        return "\r\n".join([
            f"- Host: {self.host} @ {self.port}",
            f"- Device: {self.device_name} (sn: {self.serial_number})",
            f"- Firmware version: {self.firmware_version}"
        ])

    def log_level(self, level: int):
        self.log.setLevel(level)

    @property
    def host(self) -> str:
        return self._device_info.host

    @host.setter
    def host(self, host: str):
        if not self.is_connected():
            self._device_info.host = host
        else:
            raise ConnectionError("Cannot set host when C3 is connected. Disconnect first.")

    @property
    def port(self) -> int:
        return self._device_info.port

    @port.setter
    def port(self, port: int):
        if not self.is_connected():
            self._device_info.port = port
        else:
            raise ConnectionError("Cannot set port when C3 is connected. Disconnect first.")

    @property
    def mac(self) -> str:
        return self._device_info.mac or '?'

    @property
    def serial_number(self) -> str:
        return self._device_info.serial_number or '?'

    @property
    def device_name(self) -> str:
        return self._device_info.device_name or '?'

    @property
    def firmware_version(self) -> str:
        return self._device_info.firmware_version or '?'

    @property
    def nr_of_locks(self) -> int:
        return self._status.nr_of_locks or 0

    @property
    def nr_aux_in(self) -> int:
        return self._status.nr_aux_in or 0

    @property
    def nr_aux_out(self) -> int:
        return self._status.nr_aux_out or 0

    @classmethod
    def discover(cls, interface_address: str = None, timeout: int = 2) -> list[C3]:
        """Scan on all local network interface, or the provided interface, for C3 panels."""
        devices = []
        message = cls._construct_message(None, None, consts.Command.DISCOVER, consts.C3_DISCOVERY_MESSAGE)

        if interface_address:
            ip_addresses = [interface_address]
        else:
            interfaces = socket.getaddrinfo(host=socket.gethostname(), port=None, family=socket.AF_INET)
            ip_addresses = [ip[-1][0] for ip in interfaces]
        for ip_address in ip_addresses:
            cls.log.debug("Discover on %s", ip_address)
            sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_UDP)
            sock.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
            sock.settimeout(timeout)
            sock.bind((ip_address, 0))
            sock.sendto(message, ("255.255.255.255", consts.C3_PORT_BROADCAST))

            while True:
                try:
                    payload = sock.recv(64*1024)
                except socket.timeout:
                    break

                if payload:
                    received_command, data_size, _ = cls._get_message_header(payload)
                    if received_command == consts.C3_REPLY_OK:
                        # Get the message data and signature
                        message = cls._get_message(payload)

                        if len(message) != data_size:
                            raise ValueError(
                                "Length of received message (%d) does not match specified size (%d)" % (len(message),
                                                                                                        data_size))
                        data = cls._parse_kv_from_message(message)
                        devices.append(C3(C3DeviceInfo(
                            host=data.get("IP"),
                            mac=data.get("MAC"),
                            serial_number=data.get("SN"),
                            device_name=data.get("Device"),
                            firmware_version=data.get("Ver")
                        )))
            sock.close()

        return devices

    def connect(self, password: Optional[str] = None) -> bool:
        """Connect to the C3 panel on the host/port provided in the constructor."""
        self._connected = False
        self._session_id = 0xFEFE
        self._request_nr: -258

        data = None
        if password:
            data = bytearray(password.encode('ascii'))

        # Attempt to connect to panel with session initiation command
        try:
            self._sock.connect((self._device_info.host, self._device_info.port))
            bytes_written = self._send(consts.Command.CONNECT_SESSION, data)
            if bytes_written > 0:
                receive_data, bytes_received, protocol_version = self._receive()
                if bytes_received > 2:
                    self._session_id = (receive_data[1] << 8) + receive_data[0]
                    self.log.debug("Connected with Session ID %04x", self._session_id)
                    self._session_less = False
                    self._protocol_version = protocol_version
                    self._connected = True
        except ConnectionError as ex:
            self.log.debug("Connection attempt with session to %s failed: %s", self._device_info.host, ex)
        except ValueError as ex:
            self.log.error("Reply from %s failed: %s", self._device_info.host, ex)

        # Alternatively attempt to connect to panel without session initiation
        if not self._connected:
            try:
                self._session_id = None
                self._sock.connect((self._device_info.host, self._device_info.port))
                bytes_written = self._send(consts.Command.CONNECT_SESSION_LESS, data)
                if bytes_written > 0:
                    _, _, protocol_version = self._receive()
                    self.log.debug("Connected without session")
                    self._session_less = True
                    self._protocol_version = protocol_version
                    self._connected = True
            except ConnectionError as ex:
                self.log.debug("Connection attempt without session to %s failed: %s",
                               self._device_info.host, ex)
            except ValueError as ex:
                self.log.error("Reply from %s failed: %s", self._device_info.host, ex)

        if self._connected:
            self._initialize()

        return self._connected

    def disconnect(self):
        """Disconnect from C3 panel and end session."""
        if self.is_connected():
            try:
                self._send_receive(consts.Command.DISCONNECT)
            except ConnectionError:
                # Disconnecting a broken connection should not create more trouble,
                # ignoring a ConnectionError for that reason.
                pass
            self._sock.close()

        self._connected = False
        self._session_id = None
        self._request_nr: -258

    def get_device_param(self, request_parameters: list[str]) -> dict:
        """Retrieve the requested device parameter values."""
        if self.is_connected():
            message, _ = self._send_receive(consts.Command.GETPARAM, ','.join(request_parameters))
            parameter_values = self._parse_kv_from_message(message)
        else:
            raise ConnectionError("No connection to C3 panel.")

        return parameter_values

    def _update_inout_status(self, logs: list[rtlog.RTLogRecord]):
        for log in logs:
            if isinstance(log, rtlog.DoorAlarmStatusRecord):
                for lock_nr in range(1, self.nr_of_locks+1):
                    self._set_lock_status(lock_nr, log.door_sensor_status(lock_nr), auto_close=False)

            elif isinstance(log, rtlog.EventRecord) and log.port_nr-1 in range(self.nr_of_locks):
                if log.event_type == consts.EventType.OPEN_AUX_OUTPUT:
                    self._set_aux_out_status(log.port_nr, consts.InOutStatus.OPEN, auto_close=False)
                elif log.event_type == consts.EventType.CLOSE_AUX_OUTPUT:
                    self._set_aux_out_status(log.port_nr, consts.InOutStatus.CLOSED)

                elif log.event_type == consts.EventType.OPENED_ACCIDENT:
                    # Event feedback also expected via DoorAlarmStatusRecord, handling is probably double
                    self._set_lock_status(log.port_nr, consts.InOutStatus.OPEN, auto_close=False)
                elif log.event_type == consts.EventType.DOOR_OPENED_CORRECT:
                    # Event feedback also expected via DoorAlarmStatusRecord, handling is probably double
                    self._set_lock_status(log.port_nr, consts.InOutStatus.OPEN, auto_close=False)
                elif log.event_type == consts.EventType.DOOR_CLOSED_CORRECT:
                    # Event feedback also expected via DoorAlarmStatusRecord, handling is probably double
                    self._set_lock_status(log.port_nr, consts.InOutStatus.CLOSED)

                elif log.event_type == consts.EventType.AUX_INPUT_DISCONNECT:
                    self._set_aux_in_status(log.port_nr, consts.InOutStatus.OPEN)
                elif log.event_type == consts.EventType.AUX_INPUT_SHORT:
                    self._set_aux_in_status(log.port_nr, consts.InOutStatus.CLOSED)

                elif self.door_settings(log.port_nr).sensor_type == consts.DoorSensorType.NONE:
                    # When the door has no sensor, set the status based on the lock open/close events

                    # The lock drive time is used for automatic closing
                    lock_drive_time = self.door_settings(log.port_nr).lock_drive_time

                    if log.event_type == consts.EventType.NORMAL_PUNCH_OPEN:
                        self._set_lock_status(log.port_nr, consts.InOutStatus.OPEN, auto_close=lock_drive_time)
                    # PUNCH_NORMAL_OPEN_TZ
                    # Ignore "Punch during Normal Open Time Zone", door is already open
                    # FIRST_CARD_NORMAL_OPEN
                    # Ingore "First Card Normal Open (Punch Card)", door is already open
                    elif log.event_type == consts.EventType.MULTI_CARD_OPEN:
                        self._set_lock_status(log.port_nr, consts.InOutStatus.OPEN, auto_close=lock_drive_time)
                    elif log.event_type == consts.EventType.EMERGENCY_PASS_OPEN:
                        self._set_lock_status(log.port_nr, consts.InOutStatus.OPEN, auto_close=lock_drive_time)
                    elif log.event_type == consts.EventType.OPEN_NORMAL_OPEN_TZ:
                        # Not autoclosing, door is open during normal open time zone
                        self._set_lock_status(log.port_nr, consts.InOutStatus.OPEN, auto_close=False)
                    elif log.event_type == consts.EventType.REMOTE_OPENING:
                        # Remote closing command exected for closing, not setting auto-close
                        self._set_lock_status(log.port_nr, consts.InOutStatus.OPEN, auto_close=False)
                    elif log.event_type == consts.EventType.REMOTE_CLOSING:
                        self._set_lock_status(log.port_nr, consts.InOutStatus.CLOSED)
                    elif log.event_type == consts.EventType.PRESS_FINGER_OPEN:
                        self._set_lock_status(log.port_nr, consts.InOutStatus.OPEN, auto_close=lock_drive_time)
                    elif log.event_type == consts.EventType.MULTI_CARD_OPEN_FP:
                        self._set_lock_status(log.port_nr, consts.InOutStatus.OPEN, auto_close=lock_drive_time)
                    # FP_NORMAL_OPEN_TZ
                    # Ingore "Press Fingerprint during Normal Open Time Zone", door is already open
                    elif log.event_type == consts.EventType.CARD_FP_OPEN:
                        self._set_lock_status(log.port_nr, consts.InOutStatus.OPEN, auto_close=lock_drive_time)
                    elif log.event_type == consts.EventType.FIRST_CARD_NORMAL_OPEN_FP:
                        self._set_lock_status(log.port_nr, consts.InOutStatus.OPEN, auto_close=lock_drive_time)
                    elif log.event_type == consts.EventType.FIRST_CARD_NORMAL_OPEN_CARD_FP:
                        self._set_lock_status(log.port_nr, consts.InOutStatus.OPEN, auto_close=lock_drive_time)
                    elif log.event_type == consts.EventType.DURESS_PASSWORD_OPEN:
                        self._set_lock_status(log.port_nr, consts.InOutStatus.OPEN, auto_close=lock_drive_time)
                    elif log.event_type == consts.EventType.DURESS_FP_OPEN:
                        self._set_lock_status(log.port_nr, consts.InOutStatus.OPEN, auto_close=lock_drive_time)
                    elif log.event_type == consts.EventType.EXIT_BUTTON_OPEN:
                        self._set_lock_status(log.port_nr, consts.InOutStatus.OPEN, auto_close=lock_drive_time)
                    elif log.event_type == consts.EventType.MULTI_CARD_OPEN_CARD_FP:
                        self._set_lock_status(log.port_nr, consts.InOutStatus.OPEN, auto_close=lock_drive_time)
                    elif log.event_type == consts.EventType.NORMAL_OPEN_TZ_OVER:
                        self._set_lock_status(log.port_nr, consts.InOutStatus.CLOSED)
                    elif log.event_type == consts.EventType.REMOTE_NORMAL_OPEN:
                        # Changes door to normal open, so open until normal open time ends (NORMAL_OPEN_TZ_OVER)
                        self._set_lock_status(log.port_nr, consts.InOutStatus.OPEN, auto_close=False)
                    elif log.event_type == consts.EventType.DOOR_OPEN_BY_SUPERUSER:
                        self._set_lock_status(log.port_nr, consts.InOutStatus.OPEN, auto_close=lock_drive_time)

    def get_rt_log(self) -> list[rtlog.EventRecord | rtlog.DoorAlarmStatusRecord]:
        """Retrieve the latest event or alarm records."""
        records = []

        if self.is_connected():
            message, message_length = self._send_receive(self._rtlog_command)
            if message_length:
                if self._rtlog_command == consts.Command.RTLOG_BINARY:
                    # One RT log is 16 bytes
                    # Ensure the array is not empty and a multiple of 16
                    if message_length % 16 == 0:
                        logs_messages = [message[i:i+16] for i in range(0, message_length, 16)]
                        for log_message in logs_messages:
                            self.log.debug("Received RT binary log: %s", log_message.hex())
                            records.append(rtlog.factory(log_message))
                    else:
                        # The panel firmware does not support binary mode
                        self.log.debug("Transition RT log mode to key/value")
                        self._rtlog_command = consts.Command.RTLOG_KEYVALUE
                elif self._rtlog_command == consts.Command.RTLOG_KEYVALUE:
                    kv_pairs = self._parse_kv_from_message(message)
                    self.log.debug("Received RT k/v log (%d): %s", len(kv_pairs), kv_pairs)
                    if len(kv_pairs) > 0:
                        records.append(rtlog.factory(kv_pairs))
                else:
                    raise NotImplementedError(f"The requested RT log command {self._rtlog_command} is not supported")
        else:
            raise ConnectionError("No connection to C3 panel.")

        self._update_inout_status(records)

        return records

    def _set_lock_status(self, door_nr: int, status: consts.InOutStatus, auto_close: bool | int = False) -> None:
        """Set the specified door lock status and optionally performs automatic close after specified timeout."""

        # Update the status only when status is more specific than unknown, or when no status is recorded at all
        if not status == consts.InOutStatus.UNKNOWN or door_nr not in self._status.lock_status:
            self._status.lock_status[door_nr] = status

        if status == consts.InOutStatus.OPEN and ((auto_close or 0) > 0):
            threading.Timer(auto_close, self._set_lock_status, [door_nr, consts.InOutStatus.CLOSED]).start()

    def _auto_close_lock(self, door_nr: int) -> None:
        """Set the specified door lock to closed.

        The C3 does not send an event to update the door lock activation status, only the sensor status.
        This means the lock (or alternatively the door) status is not updated for doors without sensor.
        This function is triggered by an automatic internal timer to set the lock state to closed."""
        self._status.lock_status[door_nr] = consts.InOutStatus.CLOSED

    def _set_aux_in_status(self, aux_nr: int, status: consts.InOutStatus) -> None:
        """Set the specified auxiliary input status"""

        # Update the status only when status is more specific than unknown, or when no status is recorded at all
        if not status == consts.InOutStatus.UNKNOWN or aux_nr not in self._status.aux_in_status:
            self._status.aux_in_status[aux_nr] = status

    def _set_aux_out_status(self, aux_nr: int, status: consts.InOutStatus, auto_close: bool | int = False) -> None:
        """Set the specified auxiliary output status and optionally performs automatic close after specified timeout."""

        # Update the status only when status is more specific than unknown, or when no status is recorded at all
        if not status == consts.InOutStatus.UNKNOWN or aux_nr not in self._status.aux_out_status:
            self._status.aux_out_status[aux_nr] = status

        if status == consts.InOutStatus.OPEN and ((auto_close or 0) > 0):
            threading.Timer(auto_close, self._set_aux_out_status, [aux_nr, consts.InOutStatus.CLOSED]).start()

    def _auto_close_aux_out(self, aux_nr: int) -> None:
        """Set the specified auxiliary output to closed.

        The C3 does not send an event when an auxiliary output closes after a certain duration.
        This function is triggered by an automatic internal timer to set the aux state to closed."""
        self._status.aux_out_status[aux_nr] = consts.InOutStatus.CLOSED

    def control_device(self, command: controldevice.ControlDeviceBase):
        """Send a control command to the panel."""
        if self.is_connected():
            self._send_receive(consts.Command.CONTROL, command.to_bytes())

            if isinstance(command, controldevice.ControlDeviceOutput):
                if command.operation == consts.ControlOperation.OUTPUT and command.duration < 255:
                    if command.address == consts.ControlOutputAddress.DOOR_OUTPUT and \
                            self.door_settings(command.output_number).sensor_type == consts.DoorSensorType.NONE:
                        threading.Timer(command.duration, self._auto_close_lock, [command.output_number]).start()
                    if command.address == consts.ControlOutputAddress.AUX_OUTPUT:
                        threading.Timer(command.duration, self._auto_close_aux_out, [command.output_number]).start()
        else:
            raise ConnectionError("No connection to C3 panel.")

    def door_settings(self, door_nr: int) -> C3DoorSettings:
        """Returns the settings of the door as configured on the panel"""
        if door_nr in self._status.door_settings:
            return self._status.door_settings[door_nr]
        elif not self._status.door_settings:
            for door_idx in range(self._status.nr_of_locks):
                door_prefix = f"Door{door_idx+1}"
                param_values = self.get_device_param(
                    [door_prefix + p for p in ["SensorType", "Drivertime", "Detectortime"]])
                if param_values:
                    self._status.door_settings[door_idx+1] = C3DoorSettings(
                        sensor_type=consts.DoorSensorType(int(param_values.get(door_prefix + "SensorType"))),
                        lock_drive_time=int(param_values.get(door_prefix + "Drivertime")),
                        door_alarm_timeout=int(param_values.get(door_prefix + "Detectortime")))

            return self._status.door_settings[door_nr]
        else:
            raise ValueError("Invalid door number specified (%d), 1-%d supported", door_nr, self._status.nr_of_locks)

    def lock_status(self, door_nr: int) -> consts.InOutStatus:
        """Returns the (cached) door open/close status.
        Requires a preceding call to get_rt_log to update to the latest status."""
        return self._status.lock_status[door_nr] if door_nr in self._status.lock_status else \
            consts.InOutStatus.UNKNOWN

    def aux_in_status(self, aux_nr: int) -> consts.InOutStatus:
        """Returns the (cached) auxiliary input short/disconnect status.
        Requires a preceding call to get_rt_log to update to the latest status."""
        return self._status.aux_in_status[aux_nr] if aux_nr in self._status.aux_in_status else \
            consts.InOutStatus.UNKNOWN

    def aux_out_status(self, aux_nr: int) -> consts.InOutStatus:
        """Returns the (cached) auxiliary output open/close status.
        Requires a preceding call to get_rt_log to update to the latest status."""
        return self._status.aux_out_status[aux_nr] if aux_nr in self._status.aux_out_status else \
            consts.InOutStatus.UNKNOWN



#!/usr/bin/env python3
import argparse
import logging
import sys

from c3 import C3, consts, controldevice


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('host', help='C3 panel IP address or host name')
    parser.add_argument('--password', help='Password')
    parser.add_argument('--output', choices=['door', 'aux'], required=True, help='Output is door or auxiliary')
    parser.add_argument('--number', type=int, required=True, help='Door or auxiliary output number')
    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument('--open', action="store_true", help='Set state to normal open')
    group.add_argument('--close', action="store_true", help='Set state to closed')
    parser.add_argument('--duration', type=int, help='Duration to (temporarily) open the door')
    args = parser.parse_args()

    duration = 0
    if args.open:
        duration = args.duration or 255  # Use default open when no duration is specified

    print("Connecting to %s" % args.host)
    panel = C3(args.host)
    panel.log.addHandler(logging.StreamHandler(sys.stdout))
    panel.log.setLevel(logging.DEBUG)

    try:
        if panel.connect(args.password):
            operation = controldevice.ControlDeviceOutput(args.number,
                                                          consts.ControlOutputAddress.AUX_OUTPUT if args.output == 'aux'
                                                          else consts.ControlOutputAddress.DOOR_OUTPUT, duration)
            panel.control_device(operation)
    except Exception as e:
        print(f"Parameter retrieval failed: {e}")
    finally:
        panel.disconnect()


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
import argparse
import logging
import sys
import time

from c3 import C3, rtlog


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('host', help='C3 panel IP address or host name')
    parser.add_argument('--password', help='Password')
    args = parser.parse_args()

    print("Connecting to %s" % args.host)
    panel = C3(args.host)
    panel.log.addHandler(logging.StreamHandler(sys.stdout))
    panel.log.setLevel(logging.DEBUG)

    if panel.connect(args.password):
        try:
            while True:
                last_record_is_status = False
                records = []

                if not panel.is_connected():
                    panel.connect(args.password)

                try:
                    records = panel.get_rt_log()
                except ConnectionError as ex:
                    print(f"Error retrieving RT logs: {ex}")
                    panel.disconnect()

                for record in records:
                    print(repr(record))
                    if isinstance(record, rtlog.DoorAlarmStatusRecord):
                        last_record_is_status = True

                print(f"Door status: {[repr(panel.lock_status(i+1)) for i in range(panel.nr_of_locks)]}:")
                print(f"Aux status: {[repr(panel.aux_out_status(i+1)) for i in range(panel.nr_aux_out)]}:")

                if last_record_is_status:
                    time.sleep(9)

                print("-" * 25)

                time.sleep(1)
        except KeyboardInterrupt:
            pass

    panel.disconnect()


if __name__ == "__main__":
    main()



import math
from datetime import datetime


def lsb(data):
    return data & 0x00FF


def msb(data):
    return (data >> 8) & 0x00FF


def byte_to_signed_int(byte, size=8):
    """Two's complement conversion from byte to int"""
    value = int(byte, 16) if isinstance(byte, str) else int(byte)
    if value & (1 << (size - 1)):
        value -= 1 << size
    return value


class C3DateTime(datetime):
    @classmethod
    def from_value(cls, value: int):
        """Converts a C3 time byte array in Big-Endian encoding to a datetime object
        In the C3 protocol, the time is stored in seconds as a byte array

        DateTime= ((Year-2000)*12*31 + (Month -1)*31 + (Day-1))*(24*60*60) + Hour* 60 *60 + Minute*60 + Second
        For example, the now datetime is 2010-10-26 20:54:55, so DateTime= 347748895

        And calculate the reverse “DateTime = 347748895”
         Second = DateTime  %  60
         Minute = ( DateTime / 60 ) % 60
         Hour =  ( DateTime / 3600 ) % 24
         Day = ( DateTime / 86400 )  %  31 + 1
         Month= ( DateTime / 2678400 ) % 12 + 1
         Year = (DateTime / 32140800 ) + 2000"""
        return C3DateTime(year=math.floor(value / 32140800) + 2000,
                          month=math.floor(value / 2678400) % 12 + 1,
                          day=math.floor(value / 86400) % 31 + 1,
                          hour=math.floor(value / 3600) % 24,
                          minute=math.floor(value / 60) % 60,
                          second=value % 60)

    @classmethod
    def from_str(cls, value: str):
        """Converts a datetime string like '2023-12-06 22:33:15' to a datetime object"""
        return datetime.strptime(value, "%Y-%m-%d %H:%M:%S")

    def to_value(self) -> int:
        return ((self.year - 2000) * 12 * 31 + (self.month - 1) * 31 + (self.day - 1)) * (24 * 60 * 60) + \
               (self.hour * 60 * 60) + (self.minute * 60) + self.second



#!/usr/bin/env python3
import argparse
import logging
import sys

from c3 import C3
from c3.utils import C3DateTime


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('host', help='C3 panel IP address or host name')
    parser.add_argument('--password', help='Password')
    parser.add_argument('--param', nargs='+', help='Parameter name(s) to request')
    args = parser.parse_args()

    print("Connecting to %s" % args.host)
    panel = C3(args.host)
    panel.log.addHandler(logging.StreamHandler(sys.stdout))
    panel.log.setLevel(logging.DEBUG)

    try:
        if panel.connect(args.password):
            print("Device:")
            print(repr(panel))

            param_names = args.param or ["~ZKFPVersion",
                                         "~SerialNumber",
                                         "FirmVer",
                                         "DeviceName",
                                         "LockCount",
                                         "ReaderCount",
                                         "AuxInCount",
                                         "AuxOutCount",
                                         "DateTime"]
            params = panel.get_device_param(param_names)

            for k in params:
                if k == "DateTime":
                    print("- %s: %s" % (k, C3DateTime.from_value(int(params.get(k))).isoformat()))
                else:
                    print("- %s: %s" % (k, params.get(k)))
    except Exception as e:
        print(f"Parameter retrieval failed: {e}")
    finally:
        panel.disconnect()


if __name__ == "__main__":
    main()



from c3.consts import *


def test_commands():
    connect_cmd = Command.CONNECT_SESSION
    assert 0x76 == connect_cmd
    connect_reply = C3_REPLY_OK
    assert 0xC8 == connect_reply


def test_control_operation():
    cancel_alarm = ControlOperation.CANCEL_ALARM
    assert 2 == int(cancel_alarm)
    assert "2" == str(cancel_alarm)
    assert "Cancel alarm" == cancel_alarm.description
    assert "2" == "%s" % cancel_alarm
    assert "Cancel alarm" == "%r" % cancel_alarm


def test_door_sensor_status():
    assert InOutStatus.UNKNOWN == InOutStatus(0)
    assert InOutStatus.CLOSED == InOutStatus(1)



#!/usr/bin/env python3
import argparse
import logging
import sys

from c3 import C3


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('--interface', help='IP address of the interface to look for devices on')
    args = parser.parse_args()

    C3.log.addHandler(logging.StreamHandler(sys.stdout))
    C3.log.setLevel(logging.DEBUG)
    devices = C3.discover(args.interface)
    for device in devices:
        print(f"Found device ({device.mac or '?'}):")
        print(repr(device))


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
import argparse
import logging
import sys

from c3 import C3, controldevice


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('host', help='C3 panel IP address or host name')
    parser.add_argument('--password', help='Password')
    args = parser.parse_args()

    print("Connecting to %s" % args.host)
    panel = C3(args.host)
    panel.log.addHandler(logging.StreamHandler(sys.stdout))
    panel.log.setLevel(logging.DEBUG)

    try:
        if panel.connect(args.password):
            panel.control_device(controldevice.ControlDeviceCancelAlarms())
    except Exception as e:
        print(f"Cancel alarms faied: {e}")
    finally:
        panel.disconnect()


if __name__ == "__main__":
    main()



# Ported CRC-16 from libcrc.org
# (https://github.com/lammertb/libcrc/blob/master/src/crc16.c)

CRC_POLY_16 = 0xA001
CRC_START_16 = 0x0000


class Crc16Builder:
    def __init__(self, crc: int = None):
        self._crc = crc or CRC_START_16

    @classmethod
    def _calc_divisor(cls, byte):
        poly = 0

        for _ in range(8):
            if ((poly ^ byte) & 0x0001) == 1:
                poly = (poly >> 1) ^ CRC_POLY_16
            else:
                poly = poly >> 1

            byte = byte >> 1

        return poly

    def add_byte(self, data: int):
        data = data & 0xFF        # Limit data size to byte
        crc = self._crc & 0xFFFF  # Truncate to 16bit
        msb = crc >> 8            # Take msb from 16bit crc
        self._crc = msb ^ self._calc_divisor(crc ^ data)

    @property
    def crc(self):
        return self._crc & 0xFFFF


def crc16(data, crc=None):
    builder = Crc16Builder(crc)

    if hasattr(data, '__iter__'):
        for byte in data:
            if isinstance(byte, int):
                builder.add_byte(byte)
            elif isinstance(byte, str):
                if len(byte) == 1:
                    builder.add_byte(ord(byte))
                else:
                    for char in byte:
                        builder.add_byte(ord(char))
            else:
                raise TypeError("Data of type %s is not supported" % type(byte))
    else:
        raise TypeError("Data '%s' is not iterable" % data)

    return builder.crc
